@@ -33,7 +33,7 @@
 #include <sys/mman.h>
 #include <sys/uio.h>
 
-#include <xs.h>
+#include <xenstore.h>
 #include <xenctrl.h>
 #include <xen/io/xenbus.h>
 
@@ -151,7 +151,7 @@ static void ioreq_finish(struct ioreq *ioreq)
     blkdev->requests_finished++;
 }
 
-static void ioreq_release(struct ioreq *ioreq)
+static void ioreq_release(struct ioreq *ioreq, bool finish)
 {
     struct XenBlkDev *blkdev = ioreq->blkdev;
 
@@ -159,7 +159,11 @@ static void ioreq_release(struct ioreq *ioreq)
     memset(ioreq, 0, sizeof(*ioreq));
     ioreq->blkdev = blkdev;
     LIST_INSERT_HEAD(&blkdev->freelist, ioreq, list);
-    blkdev->requests_finished--;
+    if (finish) {
+        blkdev->requests_finished--;
+    } else {
+        blkdev->requests_inflight--;
+    }
 }
 
 /*
@@ -487,7 +491,7 @@ static void blk_send_response_all(struct XenBlkDev *blkdev)
     while (!LIST_EMPTY(&blkdev->finished)) {
         ioreq = LIST_FIRST(&blkdev->finished);
 	send_notify += blk_send_response_one(ioreq);
-	ioreq_release(ioreq);
+        ioreq_release(ioreq, true);
     }
     if (send_notify)
 	xen_be_send_notify(&blkdev->xendev);
@@ -539,7 +543,7 @@ static void blk_handle_requests(struct XenBlkDev *blkdev)
         if (ioreq_parse(ioreq) != 0) {
             if (blk_send_response_one(ioreq))
                 xen_be_send_notify(&blkdev->xendev);
-            ioreq_release(ioreq);
+            ioreq_release(ioreq, false);
             continue;
         }
 
@@ -566,6 +570,15 @@ static void blk_bh(void *opaque)
     blk_handle_requests(blkdev);
 }
 
+/*
+ * We need to account for the grant allocations requiring contiguous
+ * chunks; the worst case number would be
+ *     max_req * max_seg + (max_req - 1) * (max_seg - 1) + 1,
+ * but in order to keep things simple just use
+ *     2 * max_req * max_seg.
+ */
+#define MAX_GRANTS(max_req, max_seg) (2 * (max_req) * (max_seg))
+
 static void blk_alloc(struct XenDevice *xendev)
 {
     struct XenBlkDev *blkdev = container_of(xendev, struct XenBlkDev, xendev);
@@ -576,6 +589,11 @@ static void blk_alloc(struct XenDevice *xendev)
     blkdev->bh = qemu_bh_new(blk_bh, blkdev);
     if (xen_mode != XEN_EMULATE)
         batch_maps = 1;
+    if (xc_gnttab_set_max_grants(xendev->gnttabdev,
+            MAX_GRANTS(max_requests, BLKIF_MAX_SEGMENTS_PER_REQUEST)) < 0) {
+        xen_be_printf(xendev, 0, "xc_gnttab_set_max_grants failed: %s\n",
+                      strerror(errno));
+    }
 }
 
 static int blk_init(struct XenDevice *xendev)
@@ -617,7 +635,7 @@ static int blk_init(struct XenDevice *xendev)
 	return -1;
 
     /* read-only ? */
-    qflags = BDRV_O_NOCACHE;
+    qflags = BDRV_O_CACHE_WB;
     if (strcmp(blkdev->mode, "w") == 0) {
 	mode   = O_RDWR;
 	qflags |= BDRV_O_RDWR;
